from django import forms
from taxonomy.models import AbstractNodeParent
from taxonomy import NO_PARENT



class NodeFormPartial(forms.ModelForm):
    '''
    Handle a Category and parenting. 
    It's called Partial because, like a ModelForm, it will not 
    instanciate without a Meta declaration. This could be provided 
    for example, by declaration, or through a ModelAdmin.
    '''    
    #! _errors
    # 1. It may be a modelform, but it has extra field for parent. This 
    # can be done.
    parent = forms.TypedChoiceField(
        # 2. Base default will be swapped
        coerce=lambda val: int(val),
    )
        
    def __init__(self, *args, **kwargs):
        # 3. Ensure 'initial' for purpose of setting parent fields
        if (not 'initial' in kwargs):
            kwargs['initial'] = {}
            
        # 4. dispatch between 'new' and 'change' forms
        # On either branch, followed by,
        # 5. Add choices to 'parents'
        # 6. Set a value on 'parents' (via. 'initial') to select values
        instance = kwargs.get('instance')
        if (instance):
            # probably an 'update'
            # setup parent choices
            api = self._meta.model.api(instance.id)
            self.declared_fields['parent'].choices = api.reparent_choices()
            # ...then 'select' current parent
            kwargs['initial']['parent'] = api.id_parent()
        else:
            # An 'add' or base form (used by admin)
            # setup parent choices
            self.declared_fields['parent'].choices = self._meta.model.api.initial_choices()
            # ...then 'select' current parent
            kwargs['initial']['parent'] = NO_PARENT

        super().__init__(*args, **kwargs)
        



from django.test import TestCase
from taxonomy.api import DepthNid
from ttest.models import TestCat, TestCatParent
from . import utils


# Tests the class for attributes and checks, not the action.
# ./manage.py test ttest.tests.test_api
class TestAPIBase(TestCase):
    '''
    Test API cache
    '''
    @classmethod
    def setUpTestData(cls):
        # build unparented taxonomy of eight terms
        utils.build_taxonomy(8)
        
    def setUp(self):
        self.api = TestCat.api
        
    def test_child_map(self):
        xe = self.api.child_map()
        # children includes a -1 entry for unparented
        self.assertEqual(len(xe), 9)

    def test_parent_map(self):
        xe = self.api.parent_map()
        self.assertEqual(len(xe), 8)  
             
    def test_node_map(self):
        nids = self.api.node_map()
        self.assertEqual(len(nids), 8)

    def test_ftree(self):
        tree = self.api.ftree()
        self.assertEqual(len(tree), 8)

    def test_ftree_type(self):
        tree = self.api.ftree()
        self.assertIs(tree[0].__class__, DepthNid)

    def test_tree_locations(self):
        locs = self.api.tree_locations()
        self.assertEqual(len(locs), 8)       
         
         

class TestAPIMethods(TestCase):
    @classmethod
    def setUpTestData(cls):
        # build parented taxonomy of eight terms
        utils.build_parented_taxonomy(8)
        
    def setUp(self):
        self.api = TestCat.api
        
    def test_clear(self):
        self.api.clear()
        # hidden vars avoid auto-rebuilds
        self.assertEqual(len(self.api._parents), 0)       
        self.assertEqual(len(self.api._children), 0)       
        self.assertEqual(len(self.api._nodes), 0)       
        self.assertIs(self.api._ftree, None)
        self.assertIs(self.api._tree_locations, None)

    def test_depth_id_tree(self):
        tree = self.api.depth_id_tree(max_depth=3)
        self.assertEqual(len(tree), 3)

    def test_tree(self):
        tree = self.api.tree(max_depth=5)
        self.assertEqual(len(tree), 5)
        
    #def save(new_parent_id, obj)

    def test_delete(self):
        self.api.delete()
        self.assertEqual(len(self.api._children), 0)       
                
                
                
class TestAPINodeMethods(TestCase):
    # '''
    # Test API cache
    # '''
    @classmethod
    def setUpTestData(cls):
        # build parented taxonomy of eight terms
        utils.build_parented_taxonomy(8)
        
    def setUp(self):
        self.api = TestCat.api(3)

    def test_term(self):
        t = self.api.node()
        self.assertIs(t.__class__, TestCat)
        
    def test_depth(self):
        # depth is from 0
        self.assertEqual(self.api.depth(), 2)

    def test_id_parent(self):
        self.assertEqual(self.api.id_parent(), 2)
        
    def test_id_children(self):
        self.assertEqual(len(self.api.id_children()), 1)

    def test_ascendants(self):
        # includes self
        self.assertEqual(len(self.api.ascendants()), 2)

    def test_descendants(self):
        # not includes self
        self.assertEqual(len(self.api.descendants()), 5)

    #! Can not run mutations in this test case
    # def test_delete(self):
        # self.api.delete()
        # self.assertEqual(len(TestCat.api.node_map()), 2)
    # def initial_choices(self):
    # def reparent_choices(self):
    # def depth_id_tree(self, max_depth=None):
    # def depth_id_ascendant_path(self):





# Generated by Django 3.1.1 on 2020-10-27 17:50

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='TestCat',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, help_text='Name for the category. Limited to 255 characters.', max_length=255)),
                ('weight', models.PositiveSmallIntegerField(blank=True, db_index=True, default=0, help_text='Priority for display of several categories Lower value orders first. 0 to 32767.')),
                ('slug', models.SlugField(help_text='Short name for use in urls.', max_length=64)),
                ('description', models.CharField(blank=True, default='', help_text='Description of the category. Limited to 255 characters.', max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TestCatParent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('nid', models.IntegerField(db_index=True, help_text='Category parented by another category.', verbose_name='term id')),
                ('pid', models.IntegerField(blank=True, db_index=True, default=-1, help_text='Category parenting another category.', verbose_name='parent term id')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from django.db import models
from django.core import checks
from django.urls import reverse
from taxonomy.api import NodeTreeAPI
from taxonomy.models import AbstractNode, AbstractNodeParent




class TestCat(AbstractNode):

    # Not unique. Terms may be in different taxonomies. They may
    # be duplicated at different places in a hierarchy e.g. 'sports>news'
    # 'local>news'.
    slug = models.SlugField(
        max_length=64,
        help_text="Short name for use in urls.",
    )
  
    description = models.CharField(
        max_length=255,
        blank=True,
        default='',
        help_text="Description of the category. Limited to 255 characters.",
    )
      
    def get_absolute_url(self):
        return reverse("category_detail", kwargs={"slug": self.slug})

    api = None

    def __repr__(self):
        return "Category(id:{}, name:{}, slug:{}, weight:{})".format(
            self.id,
            self.name,
            self.slug,
            self.weight,
        ) 



class TestCatParent(AbstractNodeParent):
        pass
                

# Always this format, but a new one needed for every taxonomy.
TestCat.api = NodeTreeAPI(
            TestCat, 
            TestCatParent, 
         ) 



from collections import namedtuple
from django.core.exceptions import ObjectDoesNotExist
from taxonomy import NO_PARENT, BIG_DEPTH



DepthNid = namedtuple('DepthNid', ['depth', 'nid'])
DepthNode = namedtuple('DepthNode', ['depth', 'cat'])



class ChoiceIterator:
    def __init__(self, depthNodes):
        self.depthNodes = depthNodes

    def __iter__(self):
        yield (NO_PARENT, 'None (root cat)')
        prev_depth = BIG_DEPTH
        for e in self.depthNodes:
            b = []
            depth = e.depth
            if (depth == 0):
                title = e.cat.name
            elif (prev_depth <= depth):
                title = '\u2007\u2007\u2007\u2007' * (depth - 1) + '\u2007└─\u2007' + e.cat.name
            else:
                title = '\u2007\u2007\u2007\u2007' * (depth - 1) + '\u2007\u2007\u2007\u2007' + e.cat.name
            prev_depth = depth
            yield (e.cat.id, title)


            
class NodeMethods:
    '''
    API for taxonomy models.
    '''
    def __init__(self, 
        model_node, 
        model_nodeparent, 
        cache, 
        node_id
    ):
        self.model_node = model_node 
        self.model_nodeparent = model_nodeparent 
        self.cache = cache
        self.id = node_id

    def node(self):
        return self.cache.node_map()[self.id]
        
    def depth(self):
        pos = self.cache.tree_locations().get(self.id)
        tree = self.cache.ftree()
        return tree[pos].depth

    def id_parent(self):
        pid = self.cache.parent_map()[self.id]
        if (pid == NO_PARENT):
            return None
        else:
            return pid

    def id_children(self):
        # catch NO_PARENT
        nid = self.id if (self.id > 0) else NO_PARENT
        return self.cache.child_map()[nid]
                                
    def parent(self):
        '''
        return
            Cat. If NO_PARENT, None
        '''
        pid = self.cache.parent_map()[self.id]
        if (pid == NO_PARENT):
            return None
        else:
            return self.cache.node_map()[pid]

    def children(self):
        '''
        return
            [Cat, ...]
        '''
        # catch NO_PARENT
        nid = self.id if (self.id != 0) else NO_PARENT
        node_map = self.cache.node_map()
        return [node_map[e] for e in self.cache.child_map()[nid]]

        
    def id_ascendants(self):
        '''
        Category ascendants
        The return is ordered as a single path.
        Includes the original id.
        return
            [id, ....]
        '''
        parent_map = self.cache.parent_map()
        nid = parent_map[self.id]
        b = []
        while (nid != NO_PARENT):
            b.append(nid)
            nid = parent_map[nid]
        return b
        
    def id_descendants(self, include_self=False):
        '''
        Category descendants
        The return is disorganised and does not structure for
        paths.
        Does not include the original id.
        return
            [id, ....]
        '''
        child_map = self.cache.child_map()
        stack = list.copy(child_map[self.id])
        b = []
        while (stack):
            nid = stack.pop()
            b.append(nid)
            stack.extend(list.copy(child_map[nid]))
        if (include_self):
            b.append(self.id)
        return b

    def ascendants(self):
        '''
        Disorganised collection of all ascendants.
        '''
        node_map = self.cache.node_map()
        ascendants = self.id_ascendants()
        return [ node_map[nid] for nid in ascendants ]
        
    def descendants(self, include_self=False):
        '''
        Disorganised collection of all descendants.
        '''
        node_map = self.cache.node_map()
        descendants = self.id_descendants(include_self)
        return [ node_map[nid] for nid in descendants ]
        
    def depth_id_ascendant_path(self, exclude_self=False):
        '''
        Tree ascscendants
        return
            [DepthNid, ....], ordered from root to target category.
        '''
        pos = self.cache.tree_locations().get(self.id)
        tree = self.cache.ftree()
        e = tree[pos]
        b = [e]       
        base_depth = e.depth - 1      
        while (base_depth > -1):
            pos -= 1  
            e = tree[pos]
            if (base_depth >= e.depth):
                b.append(e)
                base_depth -= 1
        b.reverse()
        if (exclude_self):
            b.pop()
        return b

    def depth_id_descendant_paths(self):
        '''
        Tree descendant paths
        Each path is a full route from the Category to a leaf,
        return
            [[DepthNid, ...], ...], ordered
        '''
        tree = self.cache.ftree()
        l = len(tree)
        pos = self.cache.tree_locations().get(self.id)
        e = tree[pos]
        base_depth = e.depth
        current_depth = base_depth
        stack = []
        
        # carries each path as it is built
        b = []
        pos += 1        
        while (pos < l):
            e = tree[pos]
            prev_depth = current_depth
            current_depth = e.depth
            if (current_depth <= base_depth):
                # i.e. if not a descendant of the given nid
                break
            if (current_depth <= prev_depth):
                # path finished, need to build new one
                stack.append(list.copy(b))
                # stem for the new path
                # needs to be relative to nid depth, as the builder path
                # is also
                b = b[:(current_depth - base_depth - 1)]
            b.append(e)
            pos += 1        
        if (b):
            stack.append(b)            
        return stack
        
    def ascendant_path(self, exclude_self=False):
        '''
        Path of Categories 
            [Cat, ....], ordered from root to target category.
        '''
        node_map = self.cache.node_map()
        path = self.depth_id_ascendant_path(exclude_self)
        return [ node_map[e.nid] for e in path ]
    
    def descentant_paths(self):
        node_map = self.cache.node_map()
        paths = self.depth_id_descendant_paths()
        return [ [node_map[e.nid] for e in path] for path in paths]

        
    def depth_id_tree(self, max_depth=None):
        '''
        Tree of depths and category ids.
        
        return
            [DepthNid, ...] depth from 9
        '''
        # depth is from main tree, not relative?
        tree = self.cache.ftree()
        l = len(tree)
        pos = self.cache.tree_locations().get(self.id)
        e = tree[pos]
        base_depth = e.depth
        max_depth = max_depth if max_depth else BIG_DEPTH
        b = [DepthNid(0, e.nid)]
        pos += 1        
        while (pos < l):
            e = tree[pos]
            reldepth = e.depth - base_depth
            if (reldepth < 1):
                # i.e. if not a descendant of the given nid
                break
            if (reldepth < max_depth):
                b.append(DepthNid(reldepth, e.nid)) 
            pos += 1        
        return b      


    def tree(self, max_depth=None):
        '''
        Tree
        return
            [DepthNode(depth, Cat)]
        '''
        node_map = self.cache.node_map()
        tree = self.depth_id_tree(max_depth)
        return [DepthNode(e.depth, node_map[e.nid]) for e in tree]
        
    def delete(self):
        '''
        Delete the Category and contents.
        Removes descendant categories and
        attached elements.
        ''' 
        descendant_nids = self.id_descendants()
        # because base Category is not in the descendants
        descendant_nids.append(self.id)
        self.model_nodeparent.objects.filter(nid__in=descendant_nids).delete()
        self.model_node.objects.filter(id__in=descendant_nids).delete()
        self.cache.clear()

    def reparent_choices(self):
        '''
        choices for a category parent update
        These are limited comared to creation. Updating only allows 
        attachment to categories that are not descendants, to avoid circular 
        references.
        '''
        #? This feels like a shambles. but the slick thing would be
        # tree ids no descendants?
        descendants = self.id_descendants()
        # add in the seed node_id. Don't want to parent on ourself
        descendants.append(self.id)
        # get them all. seems easier right now if sloooooow.
        non_descendant_id_tree = (e for e in self.cache.ftree() if (not(e.nid in descendants)))
        node_map = self.cache.node_map()
        return list(ChoiceIterator([DepthNode(e.depth, node_map[e.nid]) for e in non_descendant_id_tree]))

        
        
        
class NodeTreeAPI:
    '''
    API for taxonomy models.
    '''
    def __init__(self, 
        model_node, 
        model_nodeparent, 
    ):
        self.model_node = model_node 
        self.model_nodeparent = model_nodeparent 
    
        # cache
        self._parents = {}
        self._children = {}
        self._nodes = {}
        self._ftree = None
        self._tree_locations = None

    def generate_base_data(self):
        from django.db import connection
        nodes = {}
        parents = {}
        children = {}
        PID_Weighted_SQL = "SELECT t.*, tp.pid FROM {} tp INNER JOIN {} t ON tp.nid = t.id ORDER BY t.weight".format(
            self.model_nodeparent._meta.db_table,
            self.model_node._meta.db_table,
            )
        with connection.cursor() as cursor:
            cursor.execute(PID_Weighted_SQL)
            for r in cursor.fetchall():
                nid = r[0]
                nodes[nid] = self.model_node(*r[:-1])
                parents[nid] = r[-1]
                # Only set up children for later population.
                # Setup here ensures every nid is accounted for.
                children[nid] = []
        # and ensure NO_PARENTS entry
        children[NO_PARENT] = []
        for (nid, pid) in parents.items():
            children[pid].append(nid)
        self._nodes = nodes
        self._parents = parents
        self._children = children

    #NB self.children will always have a NO_PARENT entry unless cleared.
    def child_map(self):
        if (not self._children):
            self.generate_base_data()
        return self._children 

    def parent_map(self):
        if (not self._children):
            self.generate_base_data()
        return self._parents 

    def node_map(self):
        if (not self._children):
            self.generate_base_data()
        return self._nodes
    
    def _generate_tree(self):
        '''
        return
            [DepthNid] Tree depths are from 0...
        '''
        child_map = self.child_map() 
        stack = [iter(child_map[NO_PARENT])]
        depth = 0
        while (stack):
            depth = len(stack) - 1
            it = stack.pop()
            while(True):
                try:
                    nid = it.__next__()
                except StopIteration:
                    # exhausted. Pop a iter at a previous depth
                    break
                yield DepthNid(depth, nid)
                child_ids = child_map[nid]
                if (child_ids):
                    # append current iter, will return after processing children
                    stack.append(it)
                    # append new depth of iter
                    stack.append(iter(child_ids))
                    break

    def ftree(self):
        '''
        The tree for this taxonomy
        return
              [DepthNid, ...], ordered
        '''
        if (self._ftree is None):
            # list, because we need to run the generator, or it will
            # exhaust on multiple calls
            self._ftree = list(self._generate_tree())
        return self._ftree
    
    # index
    #IdxDepth = namedtuple('IdxDepth', ['idx', 'depth'])

    def tree_locations(self):
        if (self._tree_locations is None):
            self._tree_locations = {e.nid : idx for idx, e in enumerate(self.ftree())}
        return self._tree_locations
                 
    def clear(self):
        '''
        Empty cache for a given taxonomy
        '''
        self._parents = {}
        self._children = {}
        self._nodes = {}
        self._ftree = None
        self._tree_locations = None
        
    def __call__(self, *args, **kwargs):
        if args:
            node_id = args[0]
        else:
            node_id = self.model_node.objects.get(**kwargs).id
        return NodeMethods( 
            self.model_node, 
            self.model_nodeparent, 
            self, 
            node_id
        )

    def save(self, new_parent_id, obj):
        '''
        Save a category.
        Add necessary parentage on new category. Category parent is not checked 
        against taxonomy_id.
        '''
        self.model_node.save(obj)
        try:
            o = self.model_nodeparent.objects.get(nid=obj.id)
            o.pid = new_parent_id
        except ObjectDoesNotExist:
            o = self.model_nodeparent(pid=new_parent_id, nid=obj.id)
        self.model_nodeparent.save(o)
        self.clear()

    def delete(self):
        '''
        Delete the tree.
        Includes categories, and parentage.
        '''
        self.model_nodeparent.objects.all().delete()
        self.model_node.objects.all().delete()
        self.clear()

    def depth_id_tree(self, max_depth=None):
        if (max_depth is None):
            return list.copy(self.ftree())
        else:
            return [e for e in self.ftree() if e.depth < max_depth]

    def tree(self, max_depth=None):
        '''
        Tree
        return
            [(depth, Cat)]
        '''
        node_map = self.node_map()
        return [DepthNode(e.depth, node_map[e.nid]) for e in self.depth_id_tree(max_depth)]
                
    def initial_choices(self):
        '''
        Choices for parenting a category on creation
        '''
        return list(ChoiceIterator((e for e in self.tree())))



import math
from django.utils.safestring import mark_safe
from django.utils import html
from django.forms.utils import flatatt
from taxonomy import BIG_DEPTH



# NEW
class FlatTreeRenderer():
    '''
    Render a tree  as list items with depth as sublists
    Template needs to add outer list elements.
    '''
    sublist_attrs={}
    listitem_attrs={}
    data_attrs={}
    
    def data_template(self, data, dataattr_str):
        return data.name
            
    def rend(self, 
        tree, 
        sublist_attrs={},
        listitem_attrs={},
        data_attrs={}
        ):
        '''
        tree
            [(depth, Term)]
        '''
        self.sublist_attrs.update(sublist_attrs)
        self.listitem_attrs.update(listitem_attrs)
        self.data_attrs.update(data_attrs)
        b = []
        prev_depth = 0
        list_str = '<ul {} />'.format(flatatt(self.sublist_attrs))
        listitem_str = '<li {} />'.format(flatatt(self.listitem_attrs))
        dataattr_str = '{}'.format(flatatt(self.data_attrs))
        first_item = True
        
        for depth, data in tree:
            rend_data = self.data_template(data, dataattr_str) 
            if (prev_depth == depth):
                if (first_item == False):
                    b.append('</li>')
                else:
                    first_item = False
                b.append(listitem_str)
                b.append(rend_data)
            elif (prev_depth < depth):
                b.append(list_str)
                b.append(listitem_str)
                b.append(rend_data)
            else:
                b.append('</li>')
                i = prev_depth - depth
                while (i > 0):
                    b.append('</ul>')
                    b.append('</li>')
                    i -= 1
                b.append(listitem_str)
                b.append(rend_data)
            prev_depth = depth

        # close what we are on
        b.append('</li>')            
            
        # Need close codes for unclosed depths (likely!)
        while (prev_depth > 0):
            b.append('</ul>')
            b.append('</li>')
            prev_depth -= 1            
            
        return ''.join(b)
           
           
              
class FlatTreeRendererAsLinks(FlatTreeRenderer):
    '''
    Render a tree  as list items containing anchors, with depth as 
    sublists.
    Template needs to add outer list elements.
    '''
    url_prefix ='/category/'

    def data_template(self, 
        data,
        dataattr_str
        ):
        return '<a href="{url_prefix}{url_id}" {attrs}>{text}</a>'.format(
            url_prefix=self.url_prefix,
            #! probably not escape, but something else to protect URLs
            url_id = data.slug,
            attrs = dataattr_str,
            text= html.escape(data.name),
        )



class NodeListRenderer():
    '''
    Render a node list as list items.
    Template needs to add outer list elements.
    '''
    listitem_attrs={}
    data_attrs={}
    
    def data_template(self, data, dataattr_str):
        return data.name

    def rend(self, 
        node_list, 
        listitem_attrs={},
        data_attrs={}
        ):
        '''
        node_list
            [taxonomy_nodes]
        '''
        self.listitem_attrs.update(listitem_attrs)
        self.data_attrs.update(data_attrs)
        b = []
        listitem_str = '<li {} />'.format(flatatt(self.listitem_attrs))
        dataattr_str = '{}'.format(flatatt(self.data_attrs))
        for e in node_list:
            b.append(listitem_str)
            b.append(self.data_template(e, dataattr_str))
            b.append('</li>')
        return ''.join(b)    



class NodeListRendererAsLinks(NodeListRenderer):
    '''
    Render a node list as list items containing anchors.
    Template needs to add outer list elements.
    '''
    url_prefix ='/category/'

    def data_template(self, 
        data,
        dataattr_str
        ):
        return '<a href="{url_prefix}{url_id}" {attrs}>{text}</a>'.format(
            url_prefix=self.url_prefix,
            #! probably not escape, but something else to protect URLs
            url_id = data.slug,
            attrs = dataattr_str,
            text = html.escape(data.name),
        )
           
           
           
             
# OLD

## code-level templates
class AnchorNodeListRenderer():
    url_prefix ='/category/'

    def data_template(self, ctx):
       return '<a href="{url_prefix}{url_id}">{text}</a>'.format(
            url_prefix=ctx['url_prefix'],
            #? probably not escape, but something else to protect URLs
            url_id=ctx['url'],
            text=ctx['text'],
        )

    def get_context(self, data):
        return {
            'text': html.escape(data.name),
            'url_prefix' : self.url_prefix,
            'url' : data.slug,
        }
        
    def rend(self, node_list, klass='breadcrumb'):
        '''
        tree
            [(depth, Cat)]
        '''
        b = ['<ul class="{}">'.format(klass)]
        for e in node_list:
            b.append('<li>')
            b.append(self.data_template(self.get_context(e)))
            b.append('</li>')
        b.append('</ul>')
        return ''.join(b)    


    
#? add attributes for wrapping tags.
class FlatTreeRendererMark():
    '''
    Render a tree  as list items with depth indications
    '''
    def data_template(self, ctx):
        return ctx['text']

    def get_context(self, data):
        return {
            'text': html.escape(data.name)
            }
            
    def rend(self, tree, klass='tree'):
        '''
        tree
            [(depth, Term)]
        '''
        b = ['<ul class="{}">'.format(klass)]
        prev_depth = BIG_DEPTH

        for depth, data in tree:
            rend_data = self.data_template(self.get_context(data)) 
            if (depth == 0):
                text = rend_data
            elif (prev_depth <= depth):
                # 9492,'u25114'' 9472 '\u2500'
                text = '\u2007\u2007\u2007\u2007' * (depth - 1) + '\u2007<span>└─</span>\u2007' + rend_data
            else:
                text = '\u2007\u2007\u2007\u2007' * (depth - 1) + '\u2007\u2007\u2007\u2007' + rend_data
            prev_depth = depth
            b.append('<li>')
            b.append(text)
            b.append('</li>')
        b.append('</ul>')
        return ''.join(b)

class AnchorFlatTreeRendererMark(FlatTreeRendererMark):
    '''
    Requires the callback on the renders to deliver
    tuples [(href, text)]
    '''
    url_prefix ='/category/'

    def data_template(self, ctx):
        return '<a href="{url_prefix}{url_id}">{text}</a>'.format(
            url_prefix=ctx['url_prefix'],
            #! probably not escape, but something else to protect URLs
            url_id=ctx['url'],
            text=ctx['text'],
        )

    def get_context(self, data):
        return {
            'text': html.escape(data.name),
            'url_prefix' : self.url_prefix,
            'url' : data.slug,
        }    


   
    
class StackTreeRenderer():
    '''
    Render tree iters as a 2D graphics tree in HTML/SVG.
    
    The styles should be provides as SVG format e.g. 
    'stroke:rgb(0,220,126);stroke-width:4;'
    '''
    text_style=''
    beam_style = 'stroke:black;stroke-width:4;stroke-linecap:square;'
    stem_style = 'stroke:black;stroke-width:2;'
    
    def data_template(self, x, y, ctx):
        '''
        SVG template forr data extracted from the tree.
        Builtin assumes given data will be text.
        '''
        return ('<text x="{0}" y="{1}"style="{text_style}">{text}</text>').format(
            x, 
            y,             
            text=ctx['text'],
            text_style=self.text_style,
            )

    def get_context(self, data):
        return {
            'text': html.escape(data.name)
            }
        
    def rend_default(
            self, 
            tree, 
            x_space,
            data_height,
        ):
        '''
        Render the tree with a set of defaults based on the height of
        the data to be displayed.
        x_space
            box width in SVG units
        data_height
            A box height. Also used to calculate stem and beam
            proportions in a default
        '''
        return self.rend(
          tree, 
          x_space, 
          data_height * 4, 
          data_height * 1.5, 
          data_height * 2.5, 
          data_height,
        )

    def rend(self, 
        tree, 
        x_space, 
        y_space, 
        stem_offset,
        beam_offset,
        graphic_offset,
        ):
        '''
        Render a tree as 2D SVG.
        
        This is an approximation towards a tree as humans like to see
        a tree, as roots of a plant. As such it transposes the usual
        computer representation of a "a sequence downwards, with depth"
        to "a sequence downwards, children horizontal". Since cildren 
        expand horizontally, it uses limited horizontal space very 
        quickly.

        The tree is rendered as side-by-side boxes of x_xpace/y_space
        dimensions. The beam and stem marks are sketched in afterwards.
                
        Height and width of the finished tree are calculated on the fly
        (can not be provided as initial constraints).

        Default setup requires a tree (depth, Node) where Nodes carry 
        a 'name' field. Use 'get_context' to override data retrieval.
            
        tree
            [(depth, data)]
        @param x_space at least the largest width of data to be printed
        @param y_space at least the largest height of data to be 
        printed, plus beam/stem space. For horizontal text, 4 * text hight
        is a good start.
        @param stem_offset height above data before the stem starts. For
        horizontal text, 1.5 * text hight is a good start.
        @param beam_offset height above data for the beam. For
        horizontal text, 2.5 * text hight is a good start.
        @param graphic_offset x offset of all stem/beam graphics. For
        horizontal text, text hight is a good start.
        data_callback 
            function to extract data from each element in the tree. 
            Default returns the data itself, which assumes the data is 
            text. 
        return 
            An SVG definition, suitable for embedding in an HTML
            document. The default sizes to it's container.
            Text scales with the x,y sizes, not the webpage. It will
            resize with the image. Default data template escapes data.
        '''
        depth = 0
        ## Where items of data will be printed
        x = 0
        y = 0
        ## The dummy tag will be set later when the dimensions can be
        # calculated 
        b = ['dummy_svg_tag']
        # excessive, but I see no other way.
        depth_x_memory = [0 for x in range(len(tree))]
        prev_depth = -1
        ## track of the extent of the virtual area.
        # Note this is the left of the text area
        x_max = x
        y_max = y

        for depth, data in tree:
            y = ((depth) * y_space)
            y_max = max(y_max, y)
            if(depth > prev_depth):
                # it's a child
                # remember where prev layer was positioned
                depth_x_memory[prev_depth] = x
                # do not move x 
            # elif(depth == 0):
                # # it's a root item
                # x = x_max + x_space
                # # note the max extent
                # x_max = x
                # # needs no beam (or stem)                
            elif(depth < prev_depth):
                # it's a return to lower taxonomies (but not a root)
                # need to go one beyond current max
                x = x_max + x_space
                # note the max extent
                x_max = x
                # need a beam line from the previous position. This item
                # is always a sibling
                x_start = depth_x_memory[depth]
                b.append('<line x1="{0}" y1="{2}" x2="{1}" y2="{2}" style="{3}" />'.format(
                    x_start + graphic_offset, 
                    x + graphic_offset, 
                    y - beam_offset, 
                    self.beam_style
                ))
            else:
                # it's a sibling
                # move x along
                x_start = x
                x = x + x_space
                x_max = max(x_max, x)
                # print beam line from prev x to this
                b.append('<line x1="{0}" y1="{2}" x2="{1}" y2="{2}" style="{3}" />'.format(
                    x_start + graphic_offset, 
                    x + graphic_offset, 
                    y - beam_offset, 
                    self.beam_style
                ))
            #  print a stem line
            b.append('<line x1="{0}" y1="{1}" x2="{0}" y2="{2}" style="{3}" />'.format(
                x + graphic_offset, 
                y - stem_offset, 
                y - beam_offset, 
                self.stem_style
                ))
            # Now append the data  
            b.append(self.data_template(x, y, self.get_context(data)))   
            prev_depth = depth
        b.append('</svg>')
        # replacing 'dummy_div'
        # Need to be very tricky with the viewbox, Im afraid.
        # First x_max means left pos of the last text, so needs extra 
        # padding to the right, to cover the text itself. Also needs 
        # padding to the left, to balance that visually.
        # Next, y_max starts at 0, as depths start at zero. So a 
        # symetrical problem to x-axis, must start with padding to cover
        # the opening text. 
        # finally, to reduce the amount of padding this implies, I've 
        # halved the space, but that can not be done with y start...
        x_padding = math.floor(x_space / 2)
        y_padding = math.floor(y_space / 2)
        b[0] =  '<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" viewBox="{x_min} {y_min} {x_width} {y_max}" height="100%">'.format(
            x_min = -x_padding,
            y_min = -y_space,
            x_width=x_max + (2 * x_padding),
            y_max=y_max + (y_space + y_padding)
        )
        return ''.join(b)
  
     
      
class AnchorStackTreeRenderer(StackTreeRenderer):
    '''
    Render an SVG Stacktree with web anchors.
    Default setup requires a tree (depth, Node) where Nodes carry 'name' 
    and 'slug' fields. Use 'get_context' to override data retrieval.
    '''
    text_style='fill:cornflowerblue;'
    url_prefix ='/category/'

    def data_template(self, x, y, ctx):
        return ('<a xlink:href="{url_prefix}{url_id}"><text x="{0}" y="{1}" style="{text_style}">{text}</text></a>').format(
            x, 
            y, 
            url_prefix=ctx['url_prefix'],
            url_id=ctx['url'],
            text=ctx['text'],
            text_style=self.text_style,
        )

    def get_context(self, data):
        return {
            'text': html.escape(data.name),
            'url' : data.slug,
            'url_prefix' : self.url_prefix
        }



from django.db import models
from django.core import checks
from taxonomy.api import NodeTreeAPI
from taxonomy import NO_PARENT


class AbstractNode(models.Model):

    # A name field is required, or it creates some admin difficulties.
    # Not unique. The id is unique.
    name = models.CharField(
        max_length=255,
        blank=False,
        db_index=True,
        help_text="Name for the category. Limited to 255 characters.",
    )
        
    weight = models.PositiveSmallIntegerField(
        blank=True,
        default=0,
        db_index=True,
        help_text="Priority for display of several categories Lower value orders first. 0 to 32767.",
    )

    # Note that API usualls does CRUD actions. Admin calls api. So we
    # leave the model to do it's usual actions, in case users need those
    # for mainenance.
        
    def __repr__(self):
        return "Node(id:{}, name:{}, weight:{})".format(
            self.id,
            self.name,
            self.weight,
        )  
          
    def __str__(self):
        return "{}".format(
            self.name, 
        )

    @classmethod
    def check(cls, **kwargs):
        errors = super().check(**kwargs)
        if not cls._meta.swapped:
            if (cls.api is None):
                errors.append(
                    checks.Error(
                        "'{}' class must have an attribute 'api' declared.".format(
                            cls.__name__,
                        ),
                        id='taxonomy.E001',
                    )
                )
            if (not(isinstance(cls.api, NodeTreeAPI))):
                errors.append(
                    checks.Error(
                        "'{}' attribute 'api' must be a subclass of NodeTreeAPI.".format(
                            cls.__name__,
                        ),
                        id='taxonomy.E002',
                    )
                )
        return errors
        
    class Meta:
        abstract = True

        
        
# Subclasses are unalatered, but a new one needed for every taxonomy.                
class AbstractNodeParent(models.Model):
    '''
    Parent terms to other terms
    '''
    # Sentinel for 'pid' if unparented.
    # Now that would beggar belief, an auto-increment that allows -1...
    #NO_PARENT = -1
    
    nid = models.IntegerField(
        "term id",
        db_index=True,
        help_text="Category parented by another category.",
    )
      
    # can be self.NO_PARENT, if at root of tree
    pid = models.IntegerField(
        "parent term id",
        db_index=True,
        blank=True,
        default=NO_PARENT,
        help_text="Category parenting another category.",
    )
    
    def __repr__(self):
        return  "NodeParent(nid:{0}, pid:{1})".format(
            self.nid,
            self.pid
        ) 
        
    def __str__(self):
        return "NodeParent({}-{})".format(
            self.nid, 
            self.pid if self.pid != NO_PARENT else 'NO_PARENT', 
        )

    class Meta:
        abstract = True





from ttest.models import TestCat, TestCatParent
from taxonomy import NO_PARENT

def build_taxonomy(count):
    '''
    Make count Terms, unparented
    '''
    x = 1
    while (x <= count):
        xstr = str(x)
        obj = TestCat(
            name='term' + xstr, 
            slug='term-' + xstr, 
            description='desc' + xstr, 
            weight=3
            )
        TestCat.api.save(NO_PARENT, obj)
        x += 1

def build_parented_taxonomy(count):
    '''
    Make count Terms, parented in chain
    '''
    x = 1
    parent = NO_PARENT
    while (x <= count):
        xstr = str(x)
        obj = TestCat(
            name='term' + xstr, 
            slug='term-' + xstr, 
            description='desc' + xstr, 
            weight=3
            )
        TestCat.api.save(parent, obj)
        parent = obj.id
        x += 1



from django import template
from taxonomy.inline_templates import FlatTreeRenderer, StackTreeRenderer
from django.utils.safestring import mark_safe


register = template.Library()

# https://css-tricks.com/scale-svg/
class StackedTreeRendererNode(template.Node):
    def __init__(self, tree_data, x_space, data_height):
        self.renderer = StackTreeRenderer()
        self.tree_data = template.Variable(tree_data)
        self.x_space = int(x_space)
        self.data_height = int(data_height)
        
    def render(self, context):
        r = self.renderer.rend_default( 
            self.tree_data.resolve(context), 
            self.x_space, 
            self.data_height,
            )
        return mark_safe(r)

@register.tag        
def stacked_tree(parser, token):
    try:
        # split_contents() knows not to split quoted strings.
       tag_name, tree_data, x_space, data_height = token.split_contents()
    except ValueError:
        raise template.TemplateSyntaxError(
            "%r tag requires arguments: tree_data, x_space, data_height" % token.contents.split()[0]
        )

    return StackedTreeRendererNode(tree_data, x_space, data_height)
    


class FlatTreeRendererNode(template.Node):
    def __init__(self, tree_data):
        self.tree_data = template.Variable(tree_data)
        
    def render(self, context):
        renderer = FlatTreeRenderer()
        
        r = renderer.rend( 
            self.tree_data.resolve(context), 
            )
        return mark_safe(r)

@register.tag        
def flat_tree(parser, token):
    try:
        # split_contents() knows not to split quoted strings.
       tag_name, tree_data = token.split_contents()
    except ValueError:
        raise template.TemplateSyntaxError(
            "%r tag requires arguments: tree_data" % token.contents.split()[0]
        )

    return FlatTreeRendererNode(tree_data)


def to_kwargs(token, kwlumps):
    kwargs = {}

    for kw in kwlumps:
        k_v = kw.split("=", 1)
        if (not(len(k_v) == 2)):
            raise template.TemplateSyntaxError("Expected kw argument. tag:{{% {} %}} kw:{}".format(
               token.contents,
               kw
           ))
        k = k_v[0]
        v = k_v[1]
        if not (v[0] == v[-1] and v[0] in ('"', "'")):
           raise template.TemplateSyntaxError(
               "tag keyword argument values must be in quotes. tag:{{% {} %}} kw:{}".format(
               token.contents,
               kw
           ))
        kwargs[k] = v[1:-1]
           
    return kwargs



from django.utils import html
from django.forms.utils import flatatt

class NodeAnchorNode(template.Node):
    #! not threadsafe
    def __init__(self, node_list, url_id_field, url_stub, kwargs):
        self.node_list = template.Variable(node_list)
        self.url_id_field = url_id_field
        self.url_stub = url_stub
        self.kwargs = kwargs

    def render(self, context):
        b = ['<ul {} />'.format(flatatt(self.kwargs))]
        for e in self.node_list.resolve(context):
            b.append('<li>')
            b.append('<a href="{0}{1}">{2}</a>'.format(
                self.url_stub,
                #? probably not escape, but something else to protect URLs
                getattr(e, self.url_id_field),
                html.escape(e.name),
                )
            )
            b.append('</li>')
        b.append('</ul>')
        return mark_safe(''.join(b))

@register.tag        
def node_anchors(parser, token):
    '''
    Make anchors from a list of data (presumably Nodes).
    Looks for a 'slug' and 'name' field
    Undocumented keyword arguments are rendered as HTML attributes.

    url_prefix
        Prefix to go before the slug in the anchor URL e.g. url_prefix =
        'dvds', rende as href = '/dvds/fred_astaire' 
        Default '/category/'
    url_id_field
        Fieldname for id data for the URL. Default 'slug'
    '''
    lumps = token.split_contents()
    if(len(lumps) < 2):
        raise template.TemplateSyntaxError(
            "Breadcrumb tag needs one positional argument for data. tag:{{% {} %}}".format(
                token.contents,
            ))
    tag_name = lumps[0]
    node_list = lumps[1]
    kwargs = to_kwargs(token, lumps[2:])
    url_prefix = kwargs.pop('url_prefix', None) or '/category/'
    url_id_field = kwargs.pop('url_id_field', None) or 'slug'
    return NodeAnchorNode(node_list, url_id_field, url_prefix, kwargs)



from django.contrib import admin
from taxonomy.forms import NodeFormPartial



class NodeAdmin(admin.ModelAdmin):
    form = NodeFormPartial
    change_form_template = 'taxonomy/term_change_form.html'
    change_list_template = 'taxonomy/term_change_list.html'

    list_display = ('indented_node_titles',)
    search_fields = ['name']

        
    def indented_node_titles(self, obj):
        depth = obj.api(obj.id).depth()
        if (depth > 0):
            return '\u2007\u2007\u2007\u2007' * (depth - 1) + "\u2007●\u2007\u2007" + obj.name            
        else:
            return  obj.name
            
    indented_node_titles.short_description = 'Name'
    

    def save_model(self, request, obj, form, change):
        # usually the API would be called, but if not, this uses the 
        # API too.
        pid = form.cleaned_data.get('parent')
        obj.api.save(pid, obj)       

    def delete_model(self, request, obj):
        # usually the API would be called, but if not, this uses the 
        # API too.
        obj.api(obj.id).delete()

    def delete_queryset(self, request, queryset):
        # usually the API would be called, but if not, this uses the 
        # API too.
        #? More efficient with dedicated api.delete?
        for term in queryset:
            self.delete_model(request, term)



from django.test import TestCase
from taxonomy import NO_PARENT
from ttest.models import TestCat, TestCatParent



# Tests the class for attributes and checks, not the action.
# ./manage.py test ttest.tests.test_models
class TestModels(TestCase):
    
    def setUp(self):
        # Create a category for running tests on
        obj = TestCat.objects.create(
            name='Test1',
            weight=5,
            slug="test-1",
            description='test1',
        )
        TestCat.api.save(NO_PARENT, obj)
       
    # Think this is the only basic test?
    def test_parentage(self):
        tp = TestCatParent.objects.get(nid=1)
        self.assertEqual(tp.pid, -1)




#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['taxonomy', 'taxonomy.ttest.migrations']

package_data = \
{'': ['*'],
 'taxonomy': ['templates/*',
              'templates/taxonomy/*',
              'templatetags/*',
              'ttest/*',
              'ttest/tests/*']}

setup(name='django-taxonomy2',
      version='1.0.1',
      description='Category tree for Django. Replicable and customizable. Includes Admin. Low on features, easy setup.',
      author='robert crowther',
      author_email='rw.crowther@gmail.com',
      url='https://github.com/rcrowther/django-taxonomy',
      packages=packages,
      package_data=package_data,
     )
